#include <algorithm>
#include <atomic>
#include <cstdio>
#include <cstring>

#include <zephyr/device.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/kernel.h>
#include <zephyr/logging/log_ctrl.h>
#include <zephyr/net/http/client.h>
#include <zephyr/net/socket.h>
#include <zephyr/net/tls_credentials.h>
#include <zephyr/sys/reboot.h>

#include <date_time.h>
#include <modem/lte_lc.h>
#include <modem/modem_info.h>
#include <modem/modem_key_mgmt.h>

#include <zephyr/logging/log.h>
LOG_MODULE_REGISTER(app);

#define XSTR(s) STR(s)
#define STR(s) #s

// http parameters
#define HTTP_DELAY 3
#define HTTP_RESPONSE_BODY_SIZE 100
#define HTTP_HOSTNAME "us-central1-development-222623.cloudfunctions.net"
// clang-format off
#define HTTP_URL "/test_long_http_get?delay=" XSTR(HTTP_DELAY) "&size=" XSTR(HTTP_RESPONSE_BODY_SIZE)
// clang-format on

[[noreturn]] static void FatalError(const char *cause);
static bool CustomBoardInit();
static void WatchdogFeed();
static bool NetworkInit();
static bool NetworkLogInfo();
static bool HttpGet();

int main(void)
{
    LOG_INF("main() started");
    k_sleep(K_SECONDS(1));

    WatchdogFeed();
    if(!CustomBoardInit())
    {
        FatalError("CustomBoardInit() failed");
    }

    WatchdogFeed();
    if(!NetworkInit())
    {
        FatalError("NetworkInit() failed");
    }

    WatchdogFeed();
    if(!NetworkLogInfo())
    {
        FatalError("NetworkLogInfo() failed");
    }

    WatchdogFeed();
    if(!HttpGet())
    {
        FatalError("HttpGet() failed");
    }

    LOG_INF("HttpGet() was successful!!!");
    for(;;)
    {
        WatchdogFeed();
        k_sleep(K_SECONDS(1));
    }
}

[[noreturn]] static void FatalError(const char *cause)
{
    log_panic();
    LOG_ERR("fatal error: %s", cause);
    // normally would reset but this demo hits modem reset rate limits
    for(;;)
    {
        WatchdogFeed();
        k_sleep(K_SECONDS(1));
    }
}

static bool CustomBoardInit()
{
    LOG_INF("CustomBoardInit() started");

#ifndef CONFIG_BOARD_RETAIL_AWARE_BASE_NRF9160_NS
    LOG_INF("not using custom board, no special init needed");
#else
    // enable GPIO drivers needed on the custom board for UART trace collection
    const struct device *dev = device_get_binding("GPIO_0");
    if(dev == nullptr)
    {
        LOG_ERR("failed to get GPIO_0 device");
        return false;
    }
    if(0 != gpio_pin_configure(dev, 13, GPIO_OUTPUT_HIGH))
    {
        LOG_ERR("failed to configure pin 0.13");
        return false;
    }
    if(0 != gpio_pin_configure(dev, 14, GPIO_OUTPUT_HIGH))
    {
        LOG_ERR("failed to configure pin 0.14");
        return false;
    }
    if(0 != gpio_pin_configure(dev, 15, GPIO_OUTPUT_HIGH))
    {
        LOG_ERR("failed to configure pin 0.15");
        return false;
    }
    // allow drivers to come online
    k_sleep(K_MSEC(10));
#endif

    LOG_INF("CustomBoardInit() finished");
    return true;
}

/////////////////////////////////////////////////////////////////////////////
// watchdog

static std::atomic_uint32_t g_lastFedUptimeMs = 0;

static void WatchdogFeed()
{
    // breaks after ~50 days of uptime but is fine for simple demo
    g_lastFedUptimeMs = k_uptime_get();
}

static void WatchdogEntryPoint(void *, void *, void *)
{
    const auto sleepTime = K_MSEC(3121);
    const uint64_t feedTimeoutMs = std::max(HTTP_DELAY * 4, 60) * 1000;

    for(;;)
    {
        k_sleep(sleepTime);
        const uint64_t now = k_uptime_get();
        const uint64_t deadline = (uint64_t)g_lastFedUptimeMs + feedTimeoutMs;
        if(now > deadline)
        {
            FatalError("watchdog timed out");
        }
    }
}

K_THREAD_DEFINE(
    g_watchdogThread,
    2048,
    WatchdogEntryPoint,
    nullptr,
    nullptr,
    nullptr,
    1,
    K_USER,
    0);

/////////////////////////////////////////////////////////////////////////////
// networking

// Google cloud cert
static const char *g_cert =
    "-----BEGIN CERTIFICATE-----\n"
    "MIIDdTCCAl2gAwIBAgILBAAAAAABFUtaw5QwDQYJKoZIhvcNAQEFBQAwVzELMAkG\n"
    "A1UEBhMCQkUxGTAXBgNVBAoTEEdsb2JhbFNpZ24gbnYtc2ExEDAOBgNVBAsTB1Jv\n"
    "b3QgQ0ExGzAZBgNVBAMTEkdsb2JhbFNpZ24gUm9vdCBDQTAeFw05ODA5MDExMjAw\n"
    "MDBaFw0yODAxMjgxMjAwMDBaMFcxCzAJBgNVBAYTAkJFMRkwFwYDVQQKExBHbG9i\n"
    "YWxTaWduIG52LXNhMRAwDgYDVQQLEwdSb290IENBMRswGQYDVQQDExJHbG9iYWxT\n"
    "aWduIFJvb3QgQ0EwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDaDuaZ\n"
    "jc6j40+Kfvvxi4Mla+pIH/EqsLmVEQS98GPR4mdmzxzdzxtIK+6NiY6arymAZavp\n"
    "xy0Sy6scTHAHoT0KMM0VjU/43dSMUBUc71DuxC73/OlS8pF94G3VNTCOXkNz8kHp\n"
    "1Wrjsok6Vjk4bwY8iGlbKk3Fp1S4bInMm/k8yuX9ifUSPJJ4ltbcdG6TRGHRjcdG\n"
    "snUOhugZitVtbNV4FpWi6cgKOOvyJBNPc1STE4U6G7weNLWLBYy5d4ux2x8gkasJ\n"
    "U26Qzns3dLlwR5EiUWMWea6xrkEmCMgZK9FGqkjWZCrXgzT/LCrBbBlDSgeF59N8\n"
    "9iFo7+ryUp9/k5DPAgMBAAGjQjBAMA4GA1UdDwEB/wQEAwIBBjAPBgNVHRMBAf8E\n"
    "BTADAQH/MB0GA1UdDgQWBBRge2YaRQ2XyolQL30EzTSo//z9SzANBgkqhkiG9w0B\n"
    "AQUFAAOCAQEA1nPnfE920I2/7LqivjTFKDK1fPxsnCwrvQmeU79rXqoRSLblCKOz\n"
    "yj1hTdNGCbM+w6DjY1Ub8rrvrTnhQ7k4o+YviiY776BQVvnGCv04zcQLcFGUl5gE\n"
    "38NflNUVyRRBnMRddWQVDf9VMOyGj/8N7yy5Y0b2qvzfvGn9LhJIZJrglfCm7ymP\n"
    "AbEVtQwdpf5pLGkkeB6zpxxxYu7KyJesF12KwvhHhm4qxFYxldBniYUr+WymXUad\n"
    "DKqC5JlR3XC321Y9YeRq4VzW9v493kHMB65jUr9TU/Qr6cf9tveCX4XSQRjbgbME\n"
    "HMUfpIBvFSDJ3gyICh3WZlXi/EjJKSZp4A==\n"
    "-----END CERTIFICATE-----\n";
static constexpr nrf_sec_tag_t g_tag = 100;

static bool InitModemSecurity()
{
    const auto certLen = std::strlen(g_cert);
    if(modem_key_mgmt_cmp(
           g_tag, MODEM_KEY_MGMT_CRED_TYPE_CA_CHAIN, g_cert, certLen))
    {
        LOG_INF("Updating CA cert");
        const int result = modem_key_mgmt_write(
            g_tag, MODEM_KEY_MGMT_CRED_TYPE_CA_CHAIN, g_cert, certLen);
        if(result != 0)
        {
            LOG_ERR("modem_key_mgmt_write() failed, error: %d", result);
            return false;
        }
    }
    return true;
}

static bool InitUnixTime()
{
    for(int i = 0; i < 100; ++i)
    {
        if(date_time_is_valid())
        {
            int64_t now;
            const int result = date_time_now(&now);
            if(result != 0)
            {
                LOG_ERR("date_time_now() failed, error: %d", result);
                return false;
            }
            LOG_INF("current time: %lld", now);
            return true;
        }
        k_sleep(K_MSEC(100));
    }
    LOG_ERR("date_time_is_valid() did not report success in time");
    return false;
}

static bool NetworkInit()
{
    LOG_INF("NetworkInit() started");

    if(!InitModemSecurity())
    {
        LOG_ERR("InitModemSecurity() failed");
        return false;
    }

    const int result = lte_lc_init_and_connect();
    if(result != 0)
    {
        LOG_ERR("lte_lc_init_and_connect() failed, error: %d", result);
        return false;
    }

    if(!InitUnixTime())
    {
        LOG_ERR("InitUnixTime() failed");
        return false;
    }

    // socket connect needs retries to remove this, seems iffy without the delay
    // at least on sdk version 2.0.0 and modem firmware version 1.3.1.
    k_sleep(K_SECONDS(10));

    LOG_INF("NetworkInit() finished");
    return true;
}

static bool NetworkLogInfo()
{
    // this type is huge so we want to avoid placing it on the stack,
    // no need for thread safety in this demo as is only called from the main
    // thread
    static modem_param_info s_modemInfo;
    static bool s_modemInfoInit = false;

    if(!s_modemInfoInit)
    {
        const int result = modem_info_init();
        if(result != 0)
        {
            LOG_ERR("modem_info_init() failed, error: %d", result);
            return false;
        }
        s_modemInfoInit = true;
    }
    if(0 != modem_info_params_init(&s_modemInfo))
    {
        LOG_ERR("modem_info_params_init() failed");
        return false;
    }
    const int result = modem_info_params_get(&s_modemInfo);
    if(result != 0)
    {
        LOG_ERR("modem_info_params_get() failed, error: %d", result);
        return false;
    }

    const int signalStrength = -(int)s_modemInfo.network.rsrp.value;
    const unsigned batteryLevel = s_modemInfo.device.battery.value;
    const char *modemFirmware = s_modemInfo.device.modem_fw.value_string;

    LOG_INF("signal strength (RSRP): %d dBm", signalStrength);
    LOG_INF("battery level: %u mV", batteryLevel);
    LOG_INF("modem firmware: %s", modemFirmware);

    return true;
}

/////////////////////////////////////////////////////////////////////////////
// socket

static bool
SocketResolveHostname(const char *hostname, unsigned port, sockaddr_in &dest)
{
    std::memset(&dest, 0, sizeof(dest));

    addrinfo hints;
    addrinfo *res;
    std::memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;

    char portString[32];
    const int formatResult =
        snprintf(portString, sizeof(portString), "%u", port);
    if(formatResult < 0 || (size_t)formatResult >= sizeof(portString))
    {
        LOG_ERR("SocketResolveHostname() failed to convert port to a string");
        return false;
    }

    const int result = getaddrinfo(hostname, portString, &hints, &res);
    if(result != 0)
    {
        LOG_ERR(
            "SocketResolveHostname() failed to resolve hostname, error: %d",
            result);
        return false;
    }

    if(res != nullptr)
    {
        memcpy(&dest, res->ai_addr, res->ai_addrlen);
        freeaddrinfo(res);
        return true;
    }
    return false;
}

static bool EnableTLSVerification(
    int sock,
    const char *hostname,
    bool enableHostnameVerification)
{
    // require TLS verification
    int verify = TLS_PEER_VERIFY_REQUIRED;
    if(0 != setsockopt(sock, SOL_TLS, TLS_PEER_VERIFY, &verify, sizeof(verify)))
    {
        LOG_ERR("Failed to set socket tls verification, errno: %d", errno);
        return false;
    }

    // enable certificate authorities
    sec_tag_t tags[1] = {g_tag};
    if(0 != setsockopt(sock, SOL_TLS, TLS_SEC_TAG_LIST, tags, sizeof(tags)))
    {
        LOG_ERR("Failed to set socket tls security context, errno: %d", errno);
        return false;
    }

    if(enableHostnameVerification)
    {
        // set the hostname to use in TLS verification
        if(0
           != setsockopt(
               sock, SOL_TLS, TLS_HOSTNAME, hostname, std::strlen(hostname)))
        {
            LOG_ERR("Failed to set socket tls hostname, errno: %d", errno);
            return false;
        }
    }
    else
    {
        // temp, option will be removed
        if(0 != setsockopt(sock, SOL_TLS, TLS_HOSTNAME, nullptr, 0))
        {
            LOG_ERR(
                "Failed to disable socket tls hostname checking, errno: %d",
                errno);
            return false;
        }
    }

    return true;
}

static int SocketTCPConnect(
    const char *hostname,
    unsigned port,
    bool enableHostnameVerification = true)
{
    sockaddr_in addr;
    if(!SocketResolveHostname(hostname, port, addr))
    {
        LOG_ERR("SocketResolveHostname() failed");
        return -1;
    }

    const int sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TLS_1_2);
    if(sock < 0)
    {
        LOG_ERR("socket() failed, errno: %d", errno);
        return -1;
    }

    if(!EnableTLSVerification(sock, hostname, enableHostnameVerification))
    {
        LOG_ERR("EnableTLSVerification() failed");
        close(sock);
        return -1;
    }

    const int error = connect(
        sock, reinterpret_cast<const sockaddr *>(&addr), sizeof(sockaddr_in));
    if(error)
    {
        LOG_ERR("connect() failed, errno: %d", errno);
        close(sock);
        return -1;
    }

    LOG_INF("SocketTCPConnect() created socket %d", sock);
    return sock;
}

/////////////////////////////////////////////////////////////////////////////
// http

struct HttpUserData
{
    int statusCode = -1;
    size_t bodyLength = 0;
};

static void HttpResponseCallback(
    http_response *rsp,
    enum http_final_call final_data,
    void *user_data)
{
    LOG_INF("HttpResponseCallback() called");

    HttpUserData *data = reinterpret_cast<HttpUserData *>(user_data);

    if(data->statusCode <= 0 && rsp->http_status_code > 0)
    {
        LOG_INF("got http status code update: %d", (int)rsp->http_status_code);
        data->statusCode = rsp->http_status_code;
    }

    if(rsp->body_frag_len != 0)
    {
        LOG_INF("got a body fragment of length %u", rsp->body_frag_len);
        data->bodyLength += rsp->body_frag_len;
    }
}

static bool HttpGet()
{
    const unsigned port = 443;
    const unsigned timeoutS = HTTP_DELAY + 10;

    const int sock = SocketTCPConnect(HTTP_HOSTNAME, port, false);
    if(sock < 0)
    {
        LOG_ERR("SocketTCPConnect() failed");
        return false;
    }

    uint8_t buffer[256];
    HttpUserData userData;
    http_request request;
    std::memset(&request, 0, sizeof(request));
    request.method = HTTP_GET;
    request.url = HTTP_URL;
    request.host = HTTP_HOSTNAME;
    request.protocol = "HTTP/1.1";
    request.response = HttpResponseCallback;
    request.recv_buf = buffer;
    request.recv_buf_len = sizeof(buffer);

    const int result =
        http_client_req(sock, &request, timeoutS * 1000, &userData);
    close(sock);
    if(result < 0)
    {
        LOG_ERR("http_client_req() failed, error: %d", result);
        return false;
    }

    if(userData.statusCode != 200)
    {
        LOG_ERR(
            "server gave non-200 status code, status code: %u",
            userData.statusCode);
        return false;
    }

    if(userData.bodyLength != HTTP_RESPONSE_BODY_SIZE)
    {
        LOG_ERR(
            "server gave a response body with an unexpected size: %u",
            userData.bodyLength);
        return false;
    }

    return true;
}
